from typing import Callable, List
from lighter.config import Config
from lighter.misc import generate_long_id, DotDict


class Parameter(object):
    """
    Base Parameter class for handling hyperparameter search.
    """
    def __init__(self, ref):
        super(Parameter, self).__init__()
        config = Config.get_instance()
        if config is None:
            config = Config()
        self.config = config.copy()
        self.ref = ref
        self.compilation: List[Config] = []

    def __iter__(self):
        raise NotImplementedError('Parameter: No implementation found!')

    def __next__(self):
        raise NotImplementedError('Parameter: No implementation found!')

    def list_values(self) -> list:
        raise NotImplementedError('Parameter: No implementation found!')

    @property
    def value(self):
        return self.config.get_value(self.ref)

    @value.setter
    def value(self, value):
        self.config.set_value(self.ref, value)

    def update_config(self, config):
        self.config = config.copy()
        self.config['experiment_id'] = generate_long_id()


class GridParameter(Parameter):
    """
    Grid search parameter with linear unified step sizes.
    """
    def __init__(self, ref: str, min: float, max: float, step: float = 1.0):
        super(GridParameter, self).__init__(ref)
        self.min = min
        self.max = max
        self.step = step

    def __iter__(self):
        self.value = self.min
        return self

    def __next__(self):
        if self.value <= self.max:
            config = self.config.copy()
            self.value += self.step
            return config
        else:
            raise StopIteration

    def list_values(self) -> list:
        return list([self.min + i * self.step for i in range(int((self.max - self.min) // self.step))])


class SetParameter(Parameter):
    """
    Allows to set a single value for defining properties.
    """
    def __init__(self, ref: str, option):
        super(SetParameter, self).__init__(ref)
        self.option = option

    def __iter__(self):
        self.done = False
        return self

    def __next__(self):
        if not self.done:
            self.value = self.option
            config = self.config.copy()
            self.done = True
            return config
        else:
            raise StopIteration

    def list_values(self) -> list:
        return [self.option]


class ListParameter(Parameter):
    """
    List search parameter, which takes in a list of options and iterates them.
    """
    def __init__(self, ref: str, options: List):
        super(ListParameter, self).__init__(ref)
        self.options = options

    def __iter__(self):
        self.idx = 0
        return self

    def __next__(self):
        if self.idx < len(self.options):
            self.value = self.options[self.idx]
            config = self.config.copy()
            self.idx += 1
            return config
        else:
            raise StopIteration

    def list_values(self) -> list:
        return self.options


class CallableGridParameter(Parameter):
    """
    Grid search parameter with lambda function based step sizes which allows non linear changes.
    """
    def __init__(self, ref: str, min: float, max: float, step: Callable[[float], float]):
        super(CallableGridParameter, self).__init__(ref)
        self.min = min
        self.max = max
        self.step = step

    def __iter__(self):
        self.value = self.min
        return self

    def __next__(self):
        if self.value <= self.max:
            config = self.config.copy()
            self.value = self.step(self.value)
            return config
        else:
            raise StopIteration

    def list_values(self) -> list:
        val = self.min
        values = [val]
        while val < self.max:
            val = self.step(val)
            values.append(val)
        return values


class AnnealParameter(Parameter):
    """
    Annealing search parameter which allows lambda function non-linear decline.
    """
    def __init__(self, ref: str, start: float, threshold: float, anneal: Callable[[float], float]):
        super(AnnealParameter, self).__init__(ref)
        self.start = start
        self.threshold = threshold
        assert self.start >= self.threshold
        self.anneal = anneal

    def __iter__(self):
        self.value = self.start
        return self

    def __next__(self):
        if self.value >= self.threshold:
            config = self.config.copy()
            self.value = self.anneal(self.value)
            return config
        else:
            raise StopIteration

    def list_values(self) -> list:
        val = self.start
        values = [val]
        while val >= self.threshold:
            val = self.anneal(val)
            values.append(val)
        return values


class BinaryParameter(Parameter):
    """
    Binary search parameter which returns true and false each once.
    """
    def __init__(self, ref: str):
        super(BinaryParameter, self).__init__(ref)

    def __iter__(self):
        self.idx = 0
        self.value = False
        return self

    def __next__(self):
        if self.idx < 2:
            config = self.config.copy()
            self.value = True
            self.idx += 1
            return config
        else:
            raise StopIteration

    def list_values(self) -> list:
        return [True, False]


class StrategyParameter(Parameter):
    """
    Strategy parameters allow to iterate over different training strategies.
    """
    def __init__(self, ref: str, options: List[str], group: str = 'strategy'):
        super(StrategyParameter, self).__init__(ref)
        self.options = options
        self.group = group

    def __iter__(self):
        self.idx = 0
        return self

    def __next__(self):
        if self.idx < len(self.options):
            config = self.config.copy()
            imported_config, _ = Config.load(path=self.options[self.idx])
            for k, v in imported_config[self.group].items():
                config[self.group][k] = v
            self.idx += 1
            return config
        else:
            raise StopIteration

    def list_values(self) -> list:
        return self.options



import torch

from lighter.decorator import context, device


class BaseModule(torch.nn.Module):
    """
    Base model class descending from a PyTorch Module base class with injected 'context' instances.
    """
    @device
    @context
    def __init__(self):
        super(BaseModule, self).__init__()

    def count_parameters(self):
        return sum(p.numel() for p in self.parameters() if p.requires_grad)



import os
import shutil
from lighter.search import ParameterSearch
from lighter.decorator import context
from lighter.parameter import Parameter


class ConfigParser:
    """
    The config parser creates a set of parameters based on the decorated experiment
    class. It permutes all possible combinations within a group alias and saves the values to the
    directory.
    """
    @context
    def __init__(self, experiment=None, output_path: str = 'runs/search'):
        self.output_path = os.path.join(output_path, self.config.context_id)
        self.experiment = experiment
        if os.path.exists(self.output_path):
            shutil.rmtree(self.output_path, ignore_errors=False, onerror=None)
        os.makedirs(self.output_path)

    def parse(self, persist: bool = True):
        """
        Parses a list of configs which can be used for training different strategies.
        :param persist: bool value which allows to persist the configs to the file system
        :return: list of parameters
        """
        list_of_configs = []
        for group in self.search.keys():
            params = []
            obj = self.search[group]
            if isinstance(obj, dict):
                for key in obj.keys():
                    if issubclass(type(obj[key]), Parameter):
                        param = obj[key]
                        params.append(param)
            configs = ParameterSearch.compile(params, self.experiment.config)
            list_of_configs.extend(configs)
            if persist:
                for config in configs:
                    file_name = os.path.join(self.output_path, '{}.json'.format(config.experiment_id))
                    config.save(file_name)
        return list_of_configs



import torch
import math
import torch.nn.init as init
import torch.nn.functional as F
from torch.nn.modules.conv import _ConvNd
from torch.nn.modules.utils import _pair, _single


class Linear(torch.nn.Module):
    __constants__ = ['in_features', 'out_features']

    def __init__(self,
                 in_features,
                 out_features,
                 bias=True):
        super(Linear, self).__init__()
        self.in_features = in_features
        self.out_features = out_features

        self.weight = torch.nn.Parameter(torch.Tensor(out_features, in_features))
        if bias:
            self.bias = torch.nn.Parameter(torch.Tensor(out_features))
        else:
            self.register_parameter('bias', None)
        self.reset_parameters()

    def reset_parameters(self):
        bound = math.sqrt(3) / math.sqrt(self.in_features)
        init.uniform_(self.weight, a=-bound, b=bound)
        if self.bias is not None:
            init.constant_(self.bias, 0)

    def forward(self, input):
        weight = self.weight * torch.exp(-self.weight**2)
        bias = None
        if self.bias is not None:
            bias = self.bias * torch.exp(-self.bias**2)
        return F.linear(input, weight, bias)

    def extra_repr(self):
        return 'in_features={}, out_features={}, bias={}'.format(
            self.in_features, self.out_features, self.bias is not None
        )


class Conv2D(_ConvNd):
    def __init__(self, in_channels, out_channels, kernel_size, stride=1,
                 padding=0, dilation=1, groups=1,
                 bias=True, padding_mode='zeros'):
        kernel_size = _pair(kernel_size)
        stride = _pair(stride)
        padding = _pair(padding)
        dilation = _pair(dilation)
        super(Conv2D, self).__init__(
            in_channels, out_channels, kernel_size, stride, padding, dilation,
            False, _pair(0), groups, bias, padding_mode)

    def _conv_forward(self, input, weight, bias):
        if self.padding_mode != 'zeros':
            return F.conv2d(F.pad(input, self._reversed_padding_repeated_twice, mode=self.padding_mode),
                            weight, bias, self.stride,
                            _pair(0), self.dilation, self.groups)
        return F.conv2d(input, weight, bias, self.stride,
                        self.padding, self.dilation, self.groups)

    def forward(self, input):
        weight = self.weight * torch.exp(-self.weight**2)
        bias = self.bias
        if bias is not None:
            bias = bias * torch.exp(-bias**2)
        h = self._conv_forward(input, weight, bias)
        return h


class Conv1D(_ConvNd):
    def __init__(self, in_channels, out_channels, kernel_size, stride=1,
                 padding=0, dilation=1, groups=1,
                 bias=True, padding_mode='zeros'):
        kernel_size = _single(kernel_size)
        stride = _single(stride)
        padding = _single(padding)
        dilation = _single(dilation)
        super(Conv1D, self).__init__(
            in_channels, out_channels, kernel_size, stride, padding, dilation,
            False, _single(0), groups, bias, padding_mode)

    def _conv_forward(self, input, weight, bias):
        if self.padding_mode != 'zeros':
            return F.conv1d(F.pad(input, self._reversed_padding_repeated_twice, mode=self.padding_mode),
                            weight, bias, self.stride,
                            _single(0), self.dilation, self.groups)
        return F.conv1d(input, weight, bias, self.stride,
                        self.padding, self.dilation, self.groups)

    def forward(self, input):
        weight = self.weight * torch.exp(-self.weight**2)
        bias = self.bias
        if bias is not None:
            bias = bias * torch.exp(-bias**2)
        h = self._conv_forward(input, weight, bias)
        return h



import functools
import threading
import logging
import time
from time import sleep
from datetime import datetime, timedelta


# ==== TIMEOUT CONSTANTS ====
# timeout in seconds represents approximately 3 years
TIMEOUT_INFINITY = 99999999
TIMEOUT_24_HOURS = 86400
TIMEOUT_12_HOURS = 43200
TIMEOUT_10_HOURS = 36000
TIMEOUT_5_HOURS = 18000
TIMEOUT_2_HOURS = 7200
TIMEOUT_1_HOUR = 3600
TIMEOUT_30_MINUTES = 1800
TIMEOUT_15_MINUTES = 900
TIMEOUT_10_MINUTES = 600
TIMEOUT_5_MINUTES = 300
TIMEOUT_1_MINUTE = 60
TIMEOUT_30_SECONDS = 30
TIMEOUT_10_SECONDS = 10
TIMEOUT_5_SECONDS = 5
TIMEOUT_1_SECOND = 1


def repeated_run(func, retries=5):
    """
    Decorator for repeating processes. If an exception occurs it can retry the
    last execution.
    :param func: Function on which this decorator is applied. It was mainly meant
    for the run() method of Subprocess.
    :param retries: Maximum number of retries.
    :return: Decorator wrapper.
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        cnt = 0
        instance = args[0]
        while cnt < retries:
            try:
                func(*args, **kwargs)
                break
            except Exception as e:
                logging.warning('Proc: Process caught exception! Retrying...', e)
                cnt += 1
        if cnt >= retries:
            logging.error('Proc: Failed to complete {} execution...'.format(instance.name))
        instance._done = True
    return wrapper


class Subprocess:
    """
    Base subprocess class for all modules.
    """
    def __init__(self, name, config):
        self.thread = threading.Thread(target=self.run)
        self.config = config
        self.name = name
        self._done = False
        self._success = False
        self.timeout = None
        self.result = None
        logging.debug("Proc: {} created!".format(self.name))

    def success(self):
        return self._success

    def _complete(self):
        self._success = True
        self._done = True
        logging.info("Proc: {} done!".format(self.name))

    def run(self):
        raise NotImplementedError('Subprocess is not defined!')

    def run_async(self):
        if not self.thread.isAlive():
            self.thread.start()
        logging.debug("Proc: {} started async!".format(self.name))

    def join(self, timeout: int = None):
        self.thread.join(timeout=timeout)
        logging.debug("Proc: {} joined!".format(self.name))

    def wait(self, timeout: int = None):
        """
        Attention: Blocking start(). Waits for process to complete or timeout to
        exceed.
        :param timeout: Timeout in seconds. Attention: If not set, then
        wait is set to infinity.
        :return:
        """
        self.run_async()
        # local method caller timeout overwrites global initialized timout if not None
        if timeout is not None:
            self.timeout = timeout
        start_time = datetime.now()
        while not self._done:
            time.sleep(TIMEOUT_1_SECOND)
            if self.timeout is not None and start_time + timedelta(seconds=self.timeout) < datetime.now():
                break
        logging.debug("Proc: {} wait complete after {} sec!".format(self.name, datetime.now() - start_time))
        if self._done:
            logging.info("Proc: Successful {}!".format(self.name))
        else:
            logging.error("Proc: Failed {}!".format(self.name))


class DemoSuccessProcess(Subprocess):
    def __init__(self, name):
        super(DemoSuccessProcess, self).__init__("Demo Success Process: {}".format(name), None)

    @repeated_run
    def run(self):
        logging.info('Proc: Do some heavy computation...')
        sleep(1)
        self._complete()


class DemoFailProcess(Subprocess):
    def __init__(self, name):
        super(DemoFailProcess, self).__init__("Demo Fail Process: {}".format(name), None)
        self.first_run = 0

    @repeated_run
    def run(self):
        if self.first_run == 0:
            logging.info('Proc: Throwing exception.')
            self.first_run += 1
            raise Exception('Demo exception!')
        self._complete()


class DemoFrozenProcess(Subprocess):
    def __init__(self, name):
        super(DemoFrozenProcess, self).__init__("Demo Frozen Process: {}".format(name), None)

    @repeated_run
    def run(self):
        logging.info('Proc: This process gets stuck now... please kill me!')
        while True:
            sleep(0.1)



import socket
from datetime import datetime
from torch.utils.tensorboard import SummaryWriter
from lighter.decorator import context


class BaseWriter(SummaryWriter):
    """
    Base writer class descending from SummaryWriter (tensorboard) with injected 'context' instance.
    """
    @context
    def __init__(self, log_dir=None, comment='', purge_step=None, max_queue=10,
                 flush_secs=120, filename_suffix=''):
        if log_dir is None:
            timestamp = datetime.timestamp(datetime.now())
            log_dir = 'runs/{}-{}-{}'.format(timestamp, self.config.context_id, socket.gethostname())
        super(BaseWriter, self).__init__(log_dir=log_dir, comment=comment,
                                         purge_step=purge_step, max_queue=max_queue,
                                         flush_secs=flush_secs, filename_suffix=filename_suffix)
        self.steps = 0

    def write(self, category: str = None, *args, **kwargs):
        for key, value in kwargs.items():
            # remove collectible key prefix if available
            val = key.split('$_')
            cat, key = val[0], val[1]
            if category is not None and cat != "${}".format(category):
                continue
            # prepend category if available
            if category is not None:
                key = '{}/{}'.format(category, key)
            self.add_scalar(key, value, self.steps)

    def step(self):
        self.steps += 1



class AbortRunError(Exception):
    pass


class InvalidTypeReferenceError(Exception):
    pass


class TypeNameCollisionError(Exception):
    pass


class MultipleInstanceError(Exception):
    def __init__(self):
        super(MultipleInstanceError, self).__init__('It is not permitted to create multiple instances')


class InvalidConfigurationError(Exception):
    def __init__(self):
        super(InvalidConfigurationError, self).__init__('This configuration is not supported!')


class DependencyInjectionError(Exception):
    def __init__(self, message: str = None):
        super(DependencyInjectionError, self).__init__(
            'Object is not callable due to unresolved variable dependency. '
            'Verify your configurations and be aware that class ordering in configs matters, '
            'due to instantiation order! {}'.format(message))


class TypeInstantiationError(Exception):
    pass
